/*
*	Basic and DIGEST authentication
*	RFC2069 and RFC2617
*
*	-Kenar-
*/
#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <bio.h>

#include "httpd.h"
#include "httpsrv.h"
#include "hauth.h"
#include "misc.h"

extern int authed;

/*	ref: /sys/src/cmd/auth/factotum/httpdigest.c	*/
void
authdig1(char *user, char *realm, char *passwd, char *ha1)
{
	DigestState *s;
	uchar b[MD5dlen];
	/*
	 *  H(A1) = MD5(uid + ":" + realm ":" + passwd)
	 */
	s = md5((uchar*)user, strlen(user), nil, nil);
	md5((uchar*)":", 1, nil, s);
	md5((uchar*)realm, strlen(realm), nil, s);
	md5((uchar*)":", 1, nil, s);
	md5((uchar*)passwd, strlen(passwd), b, s);
	enc16(ha1, MD5dlen*2+1, b, MD5dlen);
	strlower(ha1);
}

/*	Ref: /sys/src/cmd/auth/factotum/httpdigest.c	*/
void
authdig2(char *ha1, char *nonce, char *method, char *uri, char *dig)
{
	uchar b[MD5dlen];
	char ha2[MD5dlen*2+1];
	DigestState *s;

	/*
	 *  H(A2) = MD5(method + ":" + uri)
	 */
	s = md5((uchar*)method, strlen(method), nil, nil);
	md5((uchar*)":", 1, nil, s);
	md5((uchar*)uri, strlen(uri), b, s);
	enc16(ha2, sizeof(ha2), b, MD5dlen);
	strlower(ha2);

	/*
	 *  digest = MD5(H(A1) + ":" + nonce + ":" + H(A2))
	 */
	s = md5((uchar*)ha1, MD5dlen*2, nil, nil);
	md5((uchar*)":", 1, nil, s);
	md5((uchar*)nonce, strlen(nonce), nil, s);
	md5((uchar*)":", 1, nil, s);
	md5((uchar*)ha2, MD5dlen*2, b, s);
	enc16(dig, MD5dlen*2+1, b, MD5dlen);
	strlower(dig);
}

static int
md5fmt(Fmt *f)
{
	char buf[MD5dlen*2+1];
	uchar *p;
	int i;

	p = va_arg(f->args, uchar*);
	for(i=0; i<MD5dlen; i++)
		sprint(buf+2*i, "%.2ux", p[i]);
	return fmtprint(f, "%s", buf);
}

/*
*	local authentication based on HTTP access authentication (RFC2617)
*
*	password file is: $web/etc/passwd
*	the format is:
*		user md5sum apath
*	or
*		user passwd apath realm
*
*	The first format is provide for compatibility to Pegasus 2.1
*	that support only basic authentication.
*	The second format is for both digest and basic authentication
*	that is provided for Pegasus 2.2.
*	Third format is for suppporting WebDAV.
*
*	Example for old format:
*		alice		54ef36ec71201fdf9d1423fd26f97f6b	/private
*	In getting the second field, be careful not to include trailing '\n'.
*	Therefore execute like this:
*		echo -n 'open sesame' | md5sum
*
*	New format contains 4 fields like below:
*		alice	f995795aebe529f93026a1388a24d22e	/private alice@hera
*	The second field comes from
*		echo -n 'alice:alice@hera:open sesamet' |md5sum
*
*	Mode bit of the passwd file must be:
*	--rw-r----- M 8 arisawa web 323 Jun 22  2006 passwd
*
*	return:
*		-1: disallowed (authentication required)
*		0:	allowed (unprotected or already authenticated)
*
*	-Kenar-
*/
int
hauth(HConnect *c, HAuth *ha, char *pwfile, char *path)
/*
*	c->head.authuser
*	c->head.authpass
*	c->head.authresponse
*	c->head.authnonce
*	c->head.authrealm
*	...
*
*	(nonce,response) is
*		(nil,nil): first time
*		(nil,passwd): response in basic authentication
*		(nonce,response): response in digest authentication
*
*	response is a raw password in case of basic authentication
*/
{
	Biobuf *b;
	char *line, *s, *old_hash, *new_hash, *huri;
	char *field[6];
	uchar digest[MD5dlen];
	char dig[2*MD5dlen+1];
	DigestState *ds;
	char buf[128]; // 128 > 2*MD5dlen + 1
	static int fmt=0;
	//char ch;
	int m, m0, n, dflag, dtype;
	char *realm, *apath;
	char *user, *passwd, *response, *nonce, *method;//, *uri, *ouri;
	char *nc, *cnonce, *qop;
	char *nextnonce;

	user = c->head.authuser;
	huri = c->head.authuri; /* used in digest authentication hash */
	passwd = c->head.authpass;
	nonce = c->head.authnonce;
	response = c->head.authresponse;
	nc = c->head.authnc;
	cnonce = c->head.authcnonce;
	qop = c->head.authqop;
	//uri = c->req.uripath;
	//ouri = c->req.ouri;
	method = c->req.meth;

	//logit(c,"DEBUG: user=%s passwd=%s response=%s", user, passwd, response);

	b = Bopen(pwfile, OREAD);
	if(b == nil){
		logit(c,"%s not open: %r",pwfile);
		return 0;
	}

	/*	we should allow all printable characters in password
	*	but currently not. this should be fixed. */
	m0 = 0;
	old_hash = nil;
	new_hash = nil;
	realm = nil;
	apath = nil;
	dflag = 0;
	dtype = 0;
	while((line = Brdline(b, '\n')) != nil){
		line[Blinelen(b)-1] = 0;
		/*	line that follow '#' is a comment */
		s = strchr(line, '#');
		if(s != nil)
			*s = '\0'; 	/* chop comment */
		n = tokenize(line, field, 6);
		if(n < 3)
			continue;
		/* then n==3 or n==4; field[2] is a protected path */
		m = deeper(path, field[2]);
		if(m > 0)
			dflag = 1;
		if(m >= m0){
			if(m > m0){
				dtype = 0;
				realm = nil;
			}
			m0 = m;
			if(!user || strcmp(user, field[0]) == 0){
				s = hstrdup(c,field[1]); /* hash value of the line */
				apath = hstrdup(c,field[2]); /* auth path */
				if(n == 3){
					/* old format. then hash is md5sum of the passwd */
					old_hash = s;
					dtype |= HAuthOldBasic;
				}
				else{
					realm = hstrdup(c,field[3]); /* realm */
					new_hash = s;
					if(n == 4)
						dtype |= (HAuthNewBasic|HAuthNewDigest);
					if(n == 5){
						s = strlower(hstrdup(c,field[4])); /* basic or digest */
						if(*s == 'b')
							dtype |= HAuthNewBasic;
						else if(*s == 'd')
							dtype |= HAuthNewDigest;
					}
				}
			}
		}
	}
	Bterm(b);

	if(!dflag) /* the path is not protected */
		return 0;


	/*	consider the case:
	*		alice	....	/foo
	*		alice	....	/foo	realm
	*	then we will have mixture of the follwings
	*		dtype 							|   realm	|   format
	*		HAuthOldBasic					|   nil		|   old
	*		HAuthNewBasic					|   !nil	|   new
	*		HAuthNewDigest					|   !nil	|   new
	*	-Kenar-	*/

	//logit(c,"DEBUG: hauth: dtype=%o user=%s apath=%s",dtype,user,apath);
	//logit(c,"DEBUG: hauth: old_hash=%s",old_hash);
	//logit(c,"DEBUG: hauth: new_hash=%s",new_hash);

	ha->dtype = dtype;
	ha->atype = HAuthNone;
	ha->path = apath;
	ha->realm = realm;
	ha->stale = 0;

	if(!user || !dtype)
		goto R;
	if(!passwd && !response)
		goto R;

	/* the client presented digest or password */
	if(passwd){
		if(realm){
			/*	then new_hash is given and the value is
			*	md5sum of user:realm:passwd 	*/
			char ha1[MD5dlen*2+1];
			authdig1(user,realm,passwd,ha1);
			if(strcmp(ha1,new_hash) == 0)
				ha->atype = HAuthBasic;
		}
		if(ha->atype == HAuthNone && old_hash){
			/* we try old format: hash is md5sum of password */
			ds = md5(nil, 0, nil, nil);
			md5((uchar*)passwd, strlen(passwd), digest, ds);
			if(fmt == 0){
				fmtinstall('M', md5fmt);
				fmt = 1;
			}
			snprint(buf, sizeof buf, "%M", digest);
			if(strcmp(old_hash, buf) == 0)
				ha->atype = HAuthBasic;
		}
		if(ha->atype == HAuthNone)
			logit(c,"hauth: fail basic user=%s realm=%s",user,realm);
	}
	else if(new_hash) { /* then we assume digest authentication */
		char buf[128];
		char *xnonce;
		/*	we use a convention:
		*		if cnonce==nil then RFC2069 digest authentication
		*		else then RFC2617 digest authentication
		*
		*	we will assign:
		*	xnonce = nonce for RFC2069
		*	xnonce = nonce + ":" + nc + ":" + cnonce + ":" + qop  for RFC2617
		*
		* 	xnonce is used for RFC2617 support. the size is probably sufficient.
		* 	note: RFC2069 and RFC2617 does not say the length of nonce value.
		*	strlen(nc) is 8. (RFC2617)
		*	strlen(nonce) is determined by server
		*	strlen(cnonce) is determined by client
		*	Safari uses 32 LHEX for cnonce that is same as response
		*	-Kenar- */

		//logit(c,"DEBUG: hauth: Digest: method=%s huri=%s",method,huri);
		xnonce = nonce;
		if(cnonce){ /* RFC2617 */
			n = snprint(buf,sizeof buf,"%s:%s:%s:%s",nonce,nc,cnonce,qop);
			/* snprint() returns strlen(buf), we examine the size */
			if(n >= sizeof(buf) - 1)
				logit(c,"hauth: so long cnonce!");
			xnonce = buf;
		}

		authdig2(new_hash, xnonce, method, huri, dig);
		//logit(c,"DEBUG: hauth: Digest: %s %s",response,dig);
		if(strcmp(response,dig) == 0) { /* response is right */
			logit(c,"hauth: OK user=%s realm=%s",user,realm);
			/* 	chknonce() checks if the nonce is registered one.
			*	if it is, returns 1 ; otherwise returns 0
			*	Suppose a browser be already authenticated in a directory
			*	and then wants to enter another directory that must be
			*	authenticated. Then the browser might use old nonce
			*	because the browser does not have new nonce.
			*	-Kenar- */
			if(cnonce && chknonce(nonce,nil) && !chknonce(nonce,response)){
				ha->atype = HAuthDigest;
				putnonce(nonce,response);
			}
			else if(!cnonce && chknonce(nonce,nil)){
				ha->atype = HAuthDigest;
				rmnonce(nonce);
			}
			else{ /* the nonce is stale or replay attack */;
				ha->stale = 1;
				logit(c,"hauth: stale, realm=%s, nonce=%s",realm,nonce);
			}
		}
		else
			logit(c,"hauth: fail digest user=%s realm=%s",user,realm);
	}
	else {
		hfail0(c, "500 Internal Error","Internal Error",
		"Error in $web/etc/passwd");
	}

R:
	/*	four types of clients will come here.
	*	0. no auth info
	*	1. basic				# passwd
	*	2. digest on RFC2069	# response
	*	3. digest on RFC2617	# response, cnonce
	*/
	if(!cnonce && !passwd){
		nextnonce = gennonce();
		putnonce(nextnonce,nil);
		c->head.nextnonce = nextnonce;
	}

	if(ha->atype){
		authed = 1;
		return 0;
	}
	return -1;
}

char *
gennonce(void) /* generate nonce value */
{
	char *e,*s;
	int i;
	char nonce[25];
	uchar randomkey[12];
	/*	- nonce is a random hexadecimal string of length 24 */
	s = nonce;
	e = s + sizeof(nonce);
	genrandom(randomkey, 12);
	for(i=0; i < 12; i++)
		s = seprint(s,e,"%02x", randomkey[i]);
	return strdup(nonce);
}


/*	nonce data base must be on a disk	*/
int
putnonce(char *nonce, char *res)
{
	char buf[80];
	int fd;
	if(res == nil){
		snprint(buf, sizeof buf,"/etc/nonce/%s", nonce);
		fd = create(buf,OREAD,DMDIR|0770);
		if(fd < 0)
			return -1;
		close(fd);
	}
	else{
		snprint(buf, sizeof buf,"/etc/nonce/%s/%s", nonce,res);
		fd = create(buf,OREAD,0660);
		if(fd < 0)
			return -1;
		close(fd);
	}
	return 0;
}

/*	remove nonce from nonce dadabase */
void
rmnonce(char *nonce)
{
	char buf[80], buf1[80];
	int fd;
	long n, i;
	Dir *dbuf;
	snprint(buf, sizeof buf,"/etc/nonce/%s", nonce);
	fd = open(buf, OREAD);
	n = dirreadall(fd,&dbuf);
	for(i = 0; i < n; i++){
		snprint(buf1, sizeof buf1,"/etc/nonce/%s/%s", nonce);
		remove(buf1);
	}
	free(dbuf);
	remove(buf);
}


/* 	returns 0, if non existent
*	returns 1, if exist
*	-Kenar- */
int
chknonce(char *nonce, char *res)
{
	char buf[80];
	if(res == nil){
		snprint(buf, sizeof buf,"/etc/nonce/%s", nonce);
		return !access(buf,AEXIST);
	}
	snprint(buf, sizeof buf,"/etc/nonce/%s/%s", nonce, res);
	return !access(buf,AEXIST);
}


/*
 * check for authorization for some parts of the server tree
 * the user name supplied with the authorization request is ignored;
 * instead, we authenticate as the realm's user.
 *
 * authorization should be done before opening any files so that
 * unauthorized users don't get to validate file names.
 *
 * returns 1 if authorized, 0 if unauthorized, -1 for io failure.
 */
int
usrchk(HConnect *c, HAuth *ha, char *pwfile, char *path)
{
	//Hio *hout;
	int status;
	char *realm;
	char *nonce;
	//int i;
	//char *s,*e;
	char buf[256];

	if(hauth(c, ha, pwfile, path) == 0)
		return 1;

	realm = ha->realm;

	if(ha->dtype & HAuthNewDigest){
		char *stale;
		char *uripath;
		//int m;

		uripath = c->req.uripath;

		if(c->head.nextnonce)
			nonce = strdup(c->head.nextnonce);
		else{
			nonce = gennonce();
			putnonce(nonce,nil); /* register the nonce value */
		}
		stale = ha->stale?"true":"false";
		/* 	stale == true	the client used old nonce value.
		*	stale != true	the client will prompted to input authentication date
		*	see rfc2069 or rfc2617 for more details */
		snprint(buf, sizeof(buf),
			"Digest realm=\"%s\", domain=\"%s\", nonce=\"%s\", stale=%s, qop=\"auth\"",
			realm,uripath,nonce,stale);
		db_put(&msgh, "WWW-Authenticate:", buf);
		//logit(c,"DEBUG: %s", buf);
		free(nonce);
		c->head.nextnonce = nil;
	}

	if(ha->dtype & (HAuthOldBasic|HAuthNewBasic)){		
		if(ha->dtype & HAuthOldBasic)
			realm = ha->path;
		snprint(buf, sizeof(buf), "Basic realm=\"%s\"", realm);
		db_add(&msgh, "WWW-Authenticate:", buf);
	}
	status = hfail(c, HUnauth, path);
	/* 0: success; -1: failure */
	if(status < 0)
		logit(c, "%s usrchk: hflush.", c->req.uripath);
	/*	Then the browser will show a login window to the user
	*	and will resend the request with some authentication
	*	information.
	*	Safari 2.0.3 is something weird. It sometimes doesn't show the login window.
	*	-Kenar- */
	return -1;
}




#ifdef XXX
void
main(int argc, char **argv)
{
	char dig[MD5dlen*2+1];
	char *user, *realm,*passwd,*nonce,*method,*uri;
	user="arisawa";
	realm="/~arisawa/";
	passwd="blackcat";
	nonce="b51f5a608f751a26919b917e";
	method="GET";
	uri="/~arisawa/";
	digest(user, realm, passwd,nonce, method, uri,dig);
	print("%s\n", dig);
	/* the result is the response that should be
	   8f18d2bfa5c3dc8bae1c1eb5ebcdaf5d  */
}
#endif
